#include <string.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "buffer.h"

void buffer_init(Buffer* bf) {
	bf->data = NULL;
	bf->data_size = 0;
}

void buffer_write(Buffer* bf, uint8_t* newdata, size_t size) {
	// избегаем переполнений
	if((bf->data_size) > 100000) buffer_clear(bf);
	
	uint8_t * b = (uint8_t*)malloc(bf->data_size+size);
	if((bf->data_size)>0) {
		memcpy(b, bf->data, bf->data_size);
		free(bf->data);
	}
	memcpy(b+(bf->data_size), newdata, size);
	bf->data_size += size;
	bf->data = b;
}

int buffer_send(Buffer* bf) {
	if((bf->data_size)==0) return 0;
	int res = SSL_write(bf->ssl, bf->data, bf->data_size);
	if(res<0) return 1;
	if(res==0) return 2;
	buffer_truncate(bf, res);
	return 0;
}

int buffer_recv(Buffer* bf) {
	uint8_t buffer[10000];
	int res = SSL_read(bf->ssl, buffer, sizeof(buffer));
	if(res<0) return 1;
	if(res==0) return 2;
	buffer_write(bf, buffer, res);
	return 0;
}

void buffer_truncate(Buffer* bf, size_t size) {
	if((bf->data_size)==0) return;
	
	if(size>=(bf->data_size)) {
		free(bf->data);
		bf->data_size = 0;
		return;
	}
	
	uint8_t * b = (uint8_t*)malloc((bf->data_size)-size);
	memcpy(b, (bf->data)+size, (bf->data_size)-size);
	free(bf->data);
	bf->data = b;
}

void buffer_clear(Buffer* bf) {
	if(bf->data_size) {
		free(bf->data);
		bf->data_size = 0;
	}
}
