#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <time.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "client_socket.h"
#include "network.h"

int cs_init(ClientSocket* cs, const char* host, uint16_t port) {
	cs->connected = 0;
	
	const SSL_METHOD *method;
	method = SSLv2_client_method();
	if(!method) {
		fprintf(stderr, "error creating method\n");
		ERR_print_errors_fp(stderr);
		return 1;
	}
	
	cs->ctx = SSL_CTX_new((SSL_METHOD *)method);
	if(!(cs->ctx)) {
		fprintf(stderr, "error creating ctx\n");
		ERR_print_errors_fp(stderr);
		return 2;
	}
	
	cs->host = malloc(strlen(host)+1);
	strcpy(cs->host, host);
	
	cs->port = port;
	
	buffer_init(&cs->br);
	buffer_init(&cs->bs);
	
	return 0;
}

int cs_connect(ClientSocket* cs) {
	cs->connected = 0;
	
	struct sockaddr_in sa;
	
	cs->sock = socket(PF_INET, SOCK_STREAM, 0);
	if((cs->sock)==-1) {
		perror("socket");
		return 1;
	}
	
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(cs->port);
	sa.sin_addr.s_addr = inet_addr(cs->host);
	
	if(connect(cs->sock, (struct sockaddr*)&sa, sizeof(sa))==-1) {
		perror("connect");
		return 2;
	}
	
	cs->ssl = SSL_new(cs->ctx);
	SSL_set_fd(cs->ssl, cs->sock);
	if(SSL_connect(cs->ssl)<0) {
		fprintf(stderr, "error connectiong through ssl\n");
		ERR_print_errors_fp(stderr);
		return 3;
	}
	
	cs->br.ssl = cs->ssl;
	cs->bs.ssl = cs->ssl;
	
	cs->connected = 1;
	
	return 0;
}

void cs_process_packet(ClientSocket* cs, uint8_t* data, uint32_t size) {
	uint8_t type = *data;
	if(type!=1) return;
	
#ifdef DEBUG
	printf("got: type=%u, size=%u\n", type, size);
#endif
	
	uint8_t* p = data+5;
	
	net_write(cs->net, p, size);
}

int cs_process(ClientSocket* cs) {
	int res = buffer_recv(&cs->br);
	if(res) {
		if(res==1) {
			fprintf(stderr, "read error\n");
		}
		if(res==2) {
			fprintf(stderr, "connection closed\n");
		}
		cs_close(cs, 1);
		return 1;
	}
	
#ifdef DEBUG
	printf("loop data_size=%u\n", cs->br.data_size);
#endif
	
	if((cs->br.data_size)<5) return 0;
	
	uint8_t* p = cs->br.data;
	
	uint32_t size = ntohl(*(uint32_t*)(p+1));
	if((cs->br.data_size)<(size+5)) return 0;
	
	cs_process_packet(cs, cs->br.data, size);
	
	buffer_truncate(&cs->br, size+5);
	return 0;
}

void cs_loop(ClientSocket* cs) {
	if(!cs->connected) {
		int res = cs_connect(cs);
		if(res) return;
		
		// flush write buffer
		cs_flush(cs);
	}
	if(cs->connected) {
		fd_set fds;
		FD_ZERO(&fds);
		
		FD_SET(cs->sock, &fds);
		
		struct timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = 500;
		int res = select(cs->sock+1, &fds, NULL, NULL, &tv);
		if(res<0) {
			if(errno==EINTR) return;
			perror("select");
			cs_close(cs, 0);
			return;
		}
		if(res==0) return;
		if(FD_ISSET(cs->sock, &fds)) (void)cs_process(cs);
	}
}

void cs_close(ClientSocket* cs, char closed) {
	if(!cs->connected) return;
	/*if(!closed) {
		SSL_shutdown(ssl);
	}*/
	SSL_free(cs->ssl);
	shutdown(cs->sock, SHUT_RDWR);
	close(cs->sock);
	
	cs->connected = 0;
}

void cs_write(ClientSocket* cs, uint8_t* data, size_t size) {
	buffer_write(&cs->bs, data, size);
	
	if(!cs->connected) return;
	
	cs_flush(cs);
}

void cs_flush(ClientSocket* cs) {
	int res = buffer_send(&cs->bs);
	if(res) {
		printf("closed2: %d\n", res);
		cs_close(cs, 1);
	}
}

