#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <openssl/ssl.h>

#include "client_socket.h"
#include "network.h"
#include "daemon.h"

char end;

void sig_handler(int signo)
{
	end = 1;
}



int main(int argc, char* argv[]) {
	const char* script = NULL;
	const char* host = "127.0.0.1";
	uint16_t port = 5678;
	int ifn = 0;
	
	if(argc>1) {
		int i = 1;
		while(i<argc) {
			if(strcmp(argv[i],"-d")==0) {
				is_daemon = 1;
				i++;
				continue;
			}
			if(strcmp(argv[i],"-k")==0) {
				daemon_kill();
				return 0;
			}
			if(strcmp(argv[i],"-n")==0) {
				if(i<(argc-1)) {
					ifn = atoi(argv[i+1]);
					i+=2;
					continue;
				}
			}
			if( (strcmp(argv[i],"-h")==0) || (strcmp(argv[i],"--help")==0) ){
				if(i<(argc-1)) {
					host = argv[i+1];
					i+=2;
					continue;
				}
			}
			if(strcmp(argv[i],"-run")==0) {
				if(i<(argc-1)) {
					script = argv[i+1];
					i+=2;
					continue;
				}
			}
			if(strcmp(argv[i],"-p")==0) {
				if(i<(argc-1)) {
					port = (uint16_t)atoi(argv[i+1]);
					i+=2;
					continue;
				}
			}
			if(strcmp(argv[i],"--help")==0) {
				printf("usage: client [-run script] [-run2 script] [-h host] [-p port] [-n tun_no] [-d] [-k]\n"
					"\t-run script  - run script after creating interface and before termination\n"
					"\t-h host      - server ip address\n"
					"\t-p port      - server port\n"
					"\t-n tun_no    - tunnel device number\n"
					"\t-d           - run as daemon\n"
					"\t-k           - terminate running daemon\n");
				printf("defaults: script=%s, host=%s, port=%u, tun_no=%d\n", script, host, port, ifn);
				return 0;
			}
			i++;
		}
	}
	
	printf("script=%s host=%s, port=%u, device=%d\n", script, host, port, ifn);
	
	srand(time(NULL));
	
	SSL_library_init();
	OpenSSL_add_all_algorithms();
	SSL_load_error_strings();
	
	ClientSocket cs;
	int res = cs_init(&cs, host, port);
	if(res) {
		printf("Can't open connection\n");
		return 1;
	}
	
	Network net;
	res = net_init(&net, ifn);
	if(res) {
		printf("Can't create network interface\n");
		return 1;
	}
	
	cs.net = &net;
	net.cs = &cs;
	
	if(script) {
		if(fork()==0) {
			printf("executing script:\n");
			execl("/bin/sh", "/bin/sh", script, net.if_name, "up", (char *) 0);
			return 0;
		}
	}
	
	printf("daemonization..\n");
	daemon_init();
	
	end = 0;
	signal(SIGINT, sig_handler);
	signal(SIGTERM, sig_handler);
	
	while(!end) {
		cs_loop(&cs);
		res = net_loop(&net);
		if(res) break;
	}
	
	cs_close(&cs, 0);
	net_close(&net);
	
	daemon_done();
	
	if(script) {
		if(fork()==0) {
			printf("executing script:\n");
			execl("/bin/sh", "/bin/sh", script, net.if_name, "down", (char *) 0);
			return 0;
		}
	}

	return 0;
}



