#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <arpa/inet.h> 
#include <sys/select.h>
#include <sys/time.h>
#include <errno.h>
#include <stdarg.h>

#ifdef __linux__
#include <linux/if.h>
#include <linux/if_tun.h>
#endif

#ifdef __FreeBSD__
#include <net/if.h>
#include <net/if_tun.h>
#endif

#include "network.h"
#include "client_socket.h"

// BUF_SIZE > MTU
#define BUF_SIZE 10000

int net_init(Network* net, int n) {
	sprintf(net->if_name, "tap%d", n);
	printf("trying to get %s\n", net->if_name);
	
#ifdef __linux__
	int flags = IFF_TAP | IFF_NO_PI;
	if(n==-1) net->if_name[0] = 0;
	
	if( (net->h = open("/dev/net/tun" , O_RDWR)) < 0 ) {
		perror("Opening /dev/net/tun");
		return 1;
	}
	
	struct ifreq ifr;
	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = flags;
	strcpy(ifr.ifr_name, net->if_name);
	
	if( (ioctl(net->h, TUNSETIFF, (void *)&ifr)) < 0 ) {
		perror("ioctl(TUNSETIFF)");
		close(net->h);
		return 2;
	}
	
	strcpy(net->if_name, ifr.ifr_name);
#else
#ifdef __FreeBSD__
	char device[100];
	sprintf(device, "/dev/%s", net->if_name);
	if( (net->h = open(device , O_RDWR)) < 0 ) {
		perror("Opening device");
		return 1;
	}
#else
#error "choose OS"
#endif
#endif
	printf("if_name=%s\n", net->if_name);
	
	printf("network opened\n");
	
	return 0;
}

int net_process(Network* net) {
#ifdef DEBUG
	printf("network::loop got something\n");
#endif
	
	uint8_t buf0[BUF_SIZE];
	int len = read(net->h, buf0, sizeof(buf0));
	if(len>0) {
		uint8_t buf[len+5];
		
		buf[0] = 1;
		uint32_t lenn = htonl(len);
		memcpy(buf+1, &lenn, sizeof(len));
		memcpy(buf+5, buf0, len);
		cs_write(net->cs, buf, len+5);
		return 0;
	} else {
		perror("network::read");
		return 1;
	}
}

int net_loop(Network* net) {
	fd_set fds;
	FD_ZERO(&fds);
	
	FD_SET(net->h, &fds);
	
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 500;
	int res = select((net->h)+1, &fds, NULL, NULL, &tv);
	if(res<0) {
		if(errno==EINTR) return 0;
		perror("select");
		return 1;
	}
	if(res==0) return 0;
	if(FD_ISSET(net->h, &fds)) {
		res = net_process(net);
		if(res) return 2;
	}
	return 0;
}

void net_close(Network* net) {
	if(!net->h) return;
	
	printf("network closed\n");
	close(net->h);
	net->h = 0;
}

void net_write(Network* net, uint8_t* data, size_t size) {
#ifdef DEBUG
	printf("network::write size=%u\n", size);
#endif
	write(net->h, data, size);
}

