#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/socket.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "client_connection.h"
#include "server.h"
#include "mac_table.h"

ClientConnection::ClientConnection(int h, struct sockaddr_in *addr, SSL_CTX *ctx, Server* server) {
	printf("new connection\n");
	this->h = h;
	this->server = server;
	memcpy(&(this->addr), addr, sizeof(struct sockaddr_in));
	
	SSL* ssl = SSL_new(ctx);
	SSL_set_fd(ssl, h);
	if(SSL_accept(ssl)<0) {
		shutdown(h, SHUT_RDWR);
		::close(h);
		
		fprintf(stderr, "error accepting ssl socket\n");
		ERR_print_errors_fp(stderr);
		throw 0;
	}
	
	bs.ssl = ssl;
	br.ssl = ssl;
}

void ClientConnection::process_packet(uint8_t* data, uint32_t size) {
	uint8_t type = *data;
	if(type!=1) return;
	
#ifdef DEBUG
	printf("======\n");
	printf("got: type=%u, size=%u from fd=%d\n", type, size, h);
#endif
	
	uint8_t* p = data+5;
	
	uint8_t dmac[MAC_SIZE];
	uint8_t smac[MAC_SIZE];
	memcpy(dmac, p, MAC_SIZE);
	memcpy(smac, p+MAC_SIZE, MAC_SIZE);
	
	server->mt.add(this, smac);
	
	bool broadcast = true;
	for(int i=0;i<MAC_SIZE;i++) {
		if(dmac[i]!=0xff) {
			broadcast = false;
			break;
		}
	}
	
	if(broadcast) {
#ifdef DEBUG
		printf("broadcasted\n");
#endif
		server->broadcast(br.data, size+5, this);
	} else {
		ClientConnection* c = server->mt.lookup(dmac);
		if(c) {
#ifdef DEBUG
			printf("sent to fd=%d\n", c->h);
#endif
			c->write(br.data, size+5);
		} else {
#ifdef DEBUG
			fprintf(stderr, "bad destination mac: %02x:%02x:%02x:%02x:%02x:%02x\n",dmac[0],dmac[1],dmac[2],dmac[3],dmac[4],dmac[5]);
#endif
		}
	}
}

void ClientConnection::loop() {
	try {
		br.recv();
	} catch(int e) {
		printf("closed: %d\n", e);
		close(true);
		return;
	}
	
#ifdef DEBUG
	printf("loop data_size=%u\n", br.data_size);
#endif
	
	if(br.data_size<5) return;
	
	uint8_t* p = br.data;
	
	uint32_t size = ntohl(*(uint32_t*)(p+1));
	if(br.data_size<(size+5)) return;
	
	process_packet(br.data, size);
	
	br.truncate(size+5);
}

void ClientConnection::close(bool closed) {
	server->mt.del(this);
	server->delete_me(this);
	
	/*if(!closed) {
		SSL_shutdown(ssl);
	}*/
	//SSL_free(ssl);
	ssl = NULL;
	shutdown(h, SHUT_RDWR);
	::close(h);
}

void ClientConnection::write(uint8_t* data, size_t size) {
	try {
		bs.write(data, size);
		bs.send();
	} catch(int e) {
		printf("closed2: %d\n", e);
		close(true);
	}
}
